""" plugin_dbui.py

	Controllers expose by the plugin:
	
		index
		database
		configuration

	Author: R. Le Gac
	$Id$

"""

__version__ = "$Revision$"


import os
import traceback
import sys

from gluon.contrib import simplejson as json
from gluon.http import HTTP
from gluon.tools import PluginManager
from pprint import pprint


def index():
	""" Main Controller to run the plugin

	SYNOPSIS
		http://localhost:8000/application/plugin_dbui[?options]

	DESCRIPTION
		Main controller to launch the plugin.
		Load the extjs librairy and the associated css files.
		Load the javascript librairy for the plugin.
		Define view.

	PUBLIC PARAMETERS WHICH CAN BE SET VIA THE PLUGINMANAGER
		
		page_view
			The page view call by the plugin
			By default view/plugin_dbui.html
			
		script_path
			The relative path where scripts are located.
			By default static/plugin_dbui/scripts
		
		with_grid_filter
			Add the javascript librairy for gridFilter
			
	PRIVATE PARAMETERS
	
		combined_fields
		field_sets
		wizards
		
	OPTIONS

		debug
			activate the debug mode on the server side.

		script foo
			run the javascript foo in the framework defined by the plugin.
			The scripts are stored in the directory application/static/script 
			by default the javascript main.js is launch.

	EXAMPLES

		http://localhost:8000/application/plugin_dbui
		http://localhost:8000/application/plugin_dbui?debug
		http://localhost:8000/application/plugin_dbui?script=foo&debug

	"""
	# check the presence of the extjs plugin
	server_path = os.path.join('applications', request.application)
	path = os.path.join(server_path, 'static', 'plugin_extjs')
	if not os.path.exists(path):
		raise HTTP(500, 'The plugin extjs is not install !')

	# is the plugin MatJax ?
	path = os.path.join(server_path, 'static', 'plugin_mathjax')
	withMathJax = (True if os.path.exists(path) else False)
	
	# option debug
	debug = "debug" in request.get_vars

	# option script
	script = (request.vars.script if "script" in request.vars else "default")
		
	if not script.endswith(".js"):
		script = '%s.js' % script

	script_path = os.path.join(server_path, plugins.dbui.script_path)
	if script not in os.listdir(script_path):
		return 'Request script "%s" does not exist !!!' % script

	# css files
	base = os.path.join(os.path.sep, request.application)
	extjs = os.path.join(base, 'static', 'plugin_extjs')
	response.files.append(os.path.join(extjs, 'resources', 'css', 'ext-all.css')) 
	response.files.append(os.path.join(extjs, 'examples', 'shared', 'icons', 'silk.css'))

	if plugins.dbui.with_grid_filter:
		cpath = os.path.join(extjs, 'examples', 'ux', 'gridfilters', 'css')
		response.files.append(os.path.join(cpath, 'GridFilters.css'))
		response.files.append(os.path.join(cpath, 'RangeMenu.css'))
		
	# javascript libraries
	if debug:
		response.files.append(os.path.join(extjs, 'adapter', 'ext', 'ext-base-debug.js'))
		response.files.append(os.path.join(extjs, 'ext-all-debug.js'))

	else:
		response.files.append(os.path.join(extjs, 'adapter', 'ext', 'ext-base.js'))
		response.files.append(os.path.join(extjs, 'ext-all.js'))

    # Ext JS extension for Ext.ux.GridFielters
	if plugins.dbui.with_grid_filter:
		jpath = os.path.join(extjs, 'examples', 'ux', 'gridfilters')
		response.files.append(os.path.join(jpath, 'menu', 'RangeMenu.js'))
		response.files.append(os.path.join(jpath, 'menu', 'ListMenu.js'))
		response.files.append(os.path.join(jpath, 'GridFilters.js'))
		response.files.append(os.path.join(jpath, 'filter', 'Filter.js'))
		response.files.append(os.path.join(jpath, 'filter', 'BooleanFilter.js'))
		response.files.append(os.path.join(jpath, 'filter', 'DateFilter.js'))
		response.files.append(os.path.join(jpath, 'filter', 'ListFilter.js'))
		response.files.append(os.path.join(jpath, 'filter', 'NumericFilter.js'))
		response.files.append(os.path.join(jpath, 'filter', 'StringFilter.js'))

	# internationalization or localization (fr, fr-fr, fr-ca, ...)
	lg = T.accepted_language
	if len(lg) == 5 and lg[0:2] == lg[3:5]: lg = lg[0:2]
	response.files.append(os.path.join(extjs, 'src', 'locale', 'ext-lang-%s.js' % lg))
	
	# plugin libraries -- respect the alphabetic order
	li = []
	appjs = os.path.join(base, 'static', 'plugin_dbui', 'lib')
	for file in os.listdir(os.path.join(server_path, 'static', 'plugin_dbui', 'lib')):
		if file.endswith(".js"):
			li.append(os.path.join(appjs, file))

	li.sort()
	response.files.extend(li)
	
	# main application script
	response.files.append(os.path.join(base, plugins.dbui.script_path, script))

	# page view
	response.view = plugins.dbui.page_view

	return dict(debug=debug, withMathJax=withMathJax)    


def database():
	""" Controllers to perform operation on the database.
	Act as a bridge between the web browser and the database service.

	All information defining the operation are sent via the POST/GET methods.
	It contains steering keyword defining the requested action.

	Return a JSON string when the action is successful.
	Otherwise return and HTTP error

	"""
	debug = "debug" in request.vars and (request.vars.debug == 'true')
	
	if debug:
		print "\nSTART DATABASE CONTROLLER"
		print "Method:", request.env.request_method
		print "Arguments:", request.vars

	try:

		# decode the data dictionary
		if "data" in request.vars and isinstance(request.vars.data, str):
			request.vars.data = json.loads(request.vars.data)
		
		di = dbSvc.proceed(request.vars)
		resp = json.dumps(di)

	except BaseException, e:
		print "Exception in the database controller:"
		print "-"*72
		traceback.print_exc(file=sys.stdout)
		print "-"*72
		raise HTTP(500, e)
		
	if debug:
		print "\nDATABASE CONTROLLER RESPONSE IS:"
		pprint(di)
		print "END OF DATABASE CONTROLLER\n"

	return resp


def configuration():
	""" Controllers/router for the web service cfgSVc
	Act as a bridge between the web browser and the configuration service.

	Relies on Ext.Direct see specification:
	http://www.extjs.com/products/extjs/direct.php

	Act as a router for the cfgSvc service (see Ext.Direct Extjs 3.1.1)
	Decode the raw HTTP post request:
	{action: "cfgSvc", method: x, data: y, tip: i, type: "rpc"}
	
	Return a JSON string with the result of the request: 
	{action: "cfgSvc", method: x, result: y, tip: i, type: "rpc"}
	
	Otherwise return and HTTP error.

	For more detail on Ext.Direct see specification:
	http://www.extjs.com/products/extjs/direct.php
	

	"""
	debug = "debug" in request.vars and (request.vars.debug == 'true')
	
	# NOTE: the server cfgSvc is start at the beginning
	# when building the model (db.py)
	
	# decode the raw HTTP post
	# ExtJs.3.1.1 {action: "xx", method: "y", data: [], type:"rpc", tid:2}
	
	req = json.loads(request.body.read())

	if debug:
		print "\nSTART CONFIGURATION CONTROLLER:" 
		pprint(req)
	
	# receive a list of dictionary for a multi-request
	li = []
	if isinstance(req, list):
		li.extend(req)
	else:
		li.append(req)
	
	for di in li:
		# construct the command to be executed
		args = di["data"]
		cmd = "%s.%s(*args)" % (di["action"], di["method"])
		
		if args == None:
			cmd = "%s.%s()" % (di["action"], di["method"])
			
		# execute the command and return HTTP exception
		# when an internal error occurred
		try:
			res = eval(cmd)
		except BaseException, e:
			print "Exception in the configuration controller:"
			print "-"*72
			traceback.print_exc(file=sys.stdout)
			print "-"*72
			raise HTTP(500, e)
	
		# NOTE: the client expect the following object:
		# ExtJs.3.1.1 {action: "xx", method: "y", result: [], type:"rpc", tid:2}
		
		del di["data"]
		di["result"] = res 

	if debug:
		print "\nCONFIGURATION CONTROLLER RESPONSE IS:"
		pprint(li)
		print "\nEND OF CONFIGURATION CONTROLLER\n"

	# encode the data as a json string
	return json.dumps(li)