""" $Id$
	Author: R. Le Gac
"""

__version__ = "$Revision$"


from foreignfield import ForeignField
from helper import encode_field
from setfield import SetField
from tools import ALL

FIELD_IN_DBFIELD = "Field %s already in the dbFields list."

# Convert a field type into a widget xtype
# The dictionary contain the configuration parameters of the Ext JS widget
FTYPE_TO_XTYPE = {'boolean':   {'xtype': 'checkbox'},\
				  'date':      {'xtype': 'datefield', 'format': 'Y-m-d'},\
				  'datetime':  {'xtype': 'datefield', 'format': 'Y-m-d H:i:s'},\
				  'double':    {'xtype': 'numberfield'},\
				  'id':   	   {'xtype': 'textfield'},\
				  'integer':   {'xtype': 'numberfield'},\
				  'password':  {'xtype': 'textfield', 'inputType': 'password'},\
				  'reference': {'xtype': 'xcombobox'},\
				  'string':    {'xtype': 'textfield'},\
				  'text':      {'xtype': 'textarea'},\
				  'time':      {'xtype': 'timefield'},\
				  'upload':    {'xtype': 'textfield', 'inputType': 'file'}}


class CfgSvc(object):

	
	def __init__(self, db, translator, pluginManager):
		
		self._db = db
		self._foreign = ForeignField(db)
		self._set = SetField(db)
		self._translate = translator
		
		for key in pluginManager.dbui:
			setattr(self, '_%s' % key, pluginManager.dbui[key])


	def _getColumn(self, tablename, fieldname):
		""" Return a dictionary with the column model 
		required by the gridPanel widget.
		
		"""
		
		col = {}
		field = self._db[tablename][fieldname]

		# column header
		col["header"] = str(self._translate(field.label))
		
		# replace foreign key by the pointing column
		if self._is_foreign_field(tablename, fieldname):
			
			k_tablename, k_fieldname, k_key = \
			self._get_foreign_data(tablename, fieldname)
			col["dataIndex"] = encode_field(k_tablename, k_fieldname)
		
		# standard column
		else:
			col["dataIndex"] = encode_field(tablename, fieldname)
			
		col["sortable"] = True
		
		# hide the primary key
		# and field which are not readable and not writable
		if fieldname == "id" or ((not field.readable) and (not field.writable)):
			col["hidden"] = True

		# Hide fields request customizing the grid
		if tablename in self._hidden_columns:
			if fieldname in self._hidden_columns[tablename]:
				col["hidden"] = True
				
		return col


	def _getColumnsModel(self, tablename):
		""" Return the columns model required by the gridPanel widget
		handling customize columns.

		"""
		
		li= []

		# standard column
		for fieldname in self._db[tablename].fields:
			li.append(self._getColumn(tablename, fieldname))
		
		# template column
		if tablename in self._template_columns:
			for el in self._template_columns[tablename]:
				
				col = {'xtype': 'templatecolumn',
					   'header': el.header,
					   'tpl': el.tpl,
					   'sortable': True}
				
				col.update(el.extjs)
				
				li.insert(el.ipos, col)
				
		return li


	def _get_foreign_data(self, tablename, fieldname):
		""" Wrapper method returning the foreign data.
		the foreign data is a tuple containing the name of the 
		pointing table, the name of the pointing field 
		and the join instruction.
		
		"""
		return self._foreign.get_foreign_data(tablename, fieldname)

	
	def _get_set_data(self, tablename, fieldname):
		""" Wrapper method
		
		"""
		return self._set.get_set_data(tablename, fieldname)
	
	
	def _getFormField(self, tablename, fieldname):
		""" Return the configuration dictionary for a form field.
		
		"""
		field = self._db[tablename][fieldname]
		cfg= {"name": encode_field(tablename, fieldname)}
					
		# foreign key
		if self._is_foreign_field(tablename, fieldname):
				
			k_tablename, k_fieldname, k_key = \
			self._get_foreign_data(tablename, fieldname)
			
			cfg["xtype"] = "xcombobox"
			cfg["hiddenName"] = encode_field(tablename, fieldname)
			cfg["displayField"] = encode_field(k_tablename, k_fieldname)
			cfg["valueField"] = encode_field(k_tablename, k_key)
			
			dbfields = [[k_tablename, k_fieldname], [k_tablename, k_key]]
			fields = [encode_field(k_tablename, k_fieldname),
					  encode_field(k_tablename, k_key)]
			
			cfg["model"] = {"store": {"table": k_tablename,
									  "dbFields": dbfields,
									  "fields": fields}}
	
		# set field
		elif self._is_set_field(tablename, fieldname):
			cfg["xtype"] = "xsetbox"
			cfg["model"] = {"setData": self._get_set_data(tablename, fieldname)}
			
		# other types of fields
		else:
			cfg.update(FTYPE_TO_XTYPE[field.type])

		# Exploit the argument of the SQLField consructor
		# field label
		cfg["fieldLabel"] = str(self._translate(field.label))

		# default value and not null
		# NOTE: the entryform doesn't work when both default and
		#       notnull condition are defined.
		if field.default:
			cfg["value"] = str(field.default)
		elif field.notnull:
			cfg["allowBlank"] = False

		# read only field
		if field.readable and not field.writable:
			cfg["readOnly"] = True

		# tool tip
		if field.comment:
			cfg["tipText"] = field.comment
			
		# hide primary key
		# and field which are not readable and not writable
		if field.name == "id" or ((not field.readable) and (not field.writable)):
			cfg["hidden"] = True
			cfg["hideLabel"] = True
			cfg["readOnly"] = True

		# customized field
		if tablename in self._customized_fields and \
		   fieldname in self._customized_fields[tablename]:
			cfg.update(self._customized_fields[tablename][fieldname])

		return cfg


	def _getFormModelItems(self, tablename):
		""" Return the list of items appearing in a Form model.
				
		"""
		li, id_is_not_used = [], True

		# form with field sets
		# NOTE: the field id is add in order to run this form
		# in all context, create, update, destroy.
		# It is add to the latest fieldset.
		
		if tablename in self._field_sets:
			for fieldset in self._field_sets[tablename]:
				
				cfg = {'title': fieldset.title, 
					   'xtype': 'fieldset',
					   'items': []}
				
				for fieldname in fieldset.fields:
					cfg['items'].append(self._getFormField(tablename, fieldname))
					if fieldname == 'id':
						id_is_not_used = False

				# protection when applying Ext JS configuration parameters
				# NOTE: if the user submit a list of items they are append at the
				# end of the current list. 
				# It is a way to add spacer and to fine tune the layout
				di = dict(fieldset.extjs)
				if 'items' in di:
					cfg['items'].extend(di['items'])
					del di['items']
				
				# Apply the remaining Ext JS configuration parameters	
				cfg.update(di)
				li.append(cfg)

			if id_is_not_used:
				li[-1]['items'].append(self._getFormField(tablename, 'id'))
					
		# the default form
		else:
			for fieldname in self._db[tablename].fields:
				li.append(self._getFormField(tablename, fieldname))
			
		return li

	
	def _getJsonStore(self, table):
		""" Generate configuration object for JSON stores 
		required by the gridPanel widget.
		
		"""

		di = {"table": table,
			  "idProperty": encode_field(table,'id'),
			  "dbFields": [],
			  "fields": []}

		for field in self._db[table].fields:
			if field in di["dbFields"]:
				raise Exception, FIELD_IN_DBFIELD 
			
			di["dbFields"].append([table, field])
			di["fields"].append(encode_field(table, field))
			
			if self._is_foreign_field(table, field):
				if "where" not in di: di["where"] = []
				k_table, k_field, k_key = self._get_foreign_data(table, field)
				
				if k_field in di["dbFields"]:
					raise Exception, FIELD_IN_DBFIELD
				
				di["dbFields"].append([k_table, k_field])
				di["fields"].append(encode_field(k_table, k_field))
				di["where"].append("[%s.%s] == [%s.%s]" % (table, field, k_table, k_key))
				
		return di
		
		
	def _getMethods(self):
		""" Return the list of methods available and 
		their number of arguments.
		
		"""
		li = []
		for el in dir(self):
			if el.startswith("_"): continue
			di = {"name": el,\
				  "len": eval("self.%s.func_code.co_argcount" % el)-1}
			li.append(di)

		return li
	

	def _is_foreign_field(self, tablename, fieldname):
		""" Wrapper method
		
		"""
		return self._foreign.is_foreign_field(tablename, fieldname)


	def _is_set_field(self, tablename, fieldname):
		""" Wrapper method
		
		"""
		return self._set.is_set_field(tablename, fieldname)

		
	def getFormModel(self, tablename):
		""" Return the configuration dictionary for an entryForm widget.
		
		"""
		model = {"title": str(self._translate(tablename.title())),
				 "table": tablename,
				 "items": self._getFormModelItems(tablename)}

		di = {"xtype": "xentry", "model": model}
		
		for el in (ALL, tablename):
			if el in self._customized_forms:
				di.update(self._customized_forms[el])
			
		return di 

		
	def getGridModel(self, tablename):
		""" Return the configuration dictionary for a gridPanel widget.
		
		"""
		
		formModel = {'xtype': 'xform',
					 'items': self._getFormModelItems(tablename)}
		
		for el in (ALL, tablename):
			if tablename in self._customized_forms:
				formModel.update(self._customized_forms[tablename])
					
		model = {}	
		model["colModel"] = self._getColumnsModel(tablename)
		model["formModel"] = formModel
		model["store"] = self._getJsonStore(tablename)
		model["table"] = tablename
		model["title"] = str(self._translate(tablename.title()))

		di = {"xtype": "xgrid", "model": model}

		for el in (ALL, tablename):
			if el in self._customized_grids:
				di.update(self._customized_grids[el])

		return di


	def getTables(self, *args):
		""" Return a dictionary with list of tables 
		and a list with the translation of their names
		
		"""
		li_1 = [el for el in self._db.tables if el not in self._hidden_tables]			
		li_1.sort()
		li_2 = [str(self._translate(el)) for el in li_1]
			
		return {'tables': li_1, 'tablesText': li_2}