""" $Id$
	Author: R. Le Gac
"""

__version__ = "$Revision$"


from basesvc import BaseSvc, KEYWORD_MISSING
from gluon.contrib import simplejson as json
from gluon.http import HTTP
from helper import encode_field, decode_field, rows_serializer
from pprint import pprint

ACTIONS = ('create', 'destroy', 'read', 'update')

# see gluon.dal.Field
FUNC_FIELDS = ('count',
			   'day',
			   'lower',
			   'max',
			   'min',
			   'minutes',
			   'month',
			   'upper',
			   'seconds',
			   'sum',
			   'year')

ACTION_INVALID = "Requested action '%s' is not valid."
FIELD_NOT_IN_DB = "the field '%s' doesn't exist in the database."
FUNC_FIELD_INVALID = "Requested function '%s' is not valid."
RECORD_DELETED = "Record %s is deleted."
RECORD_INSERTED = "Record is inserted with id %i."
RECORD_UPDATED = "Record %s is updated."
TABLE_NOT_IN_DB = "The table '%s' doesn't exist in the database."
WHERE_INVALID = "Invalid syntax for a where keywords."



class DbSvc(BaseSvc):
	""" DbSvc

	Interface the database and database model with http request.
	It is the glue between wep2py and javascript.

	Handle HTTP POST and HTTP GET request.
	The selection of the action and its parameters are based on
	steering keywords:  xaction, table, dbFields, debug,...
	The xaction is mandatory.

	Main actions are:

		create
			create a new record in a table.

		read
			read the content of a single table
			and handle query with foreign key and where conditions

		select
			read the content of several table handling complex
			query with join, group by, order by keywords.
			
		update
			update the record of a table.

		destroy
			delete the record of a table.
	
	The steering keyword table is mandatory for the actions
	create, read, update and destroy.
	
	the steering keyword for the select operation is xaction=read
	but without table definition.
	
	Each action return a dictionary:
	{success: True, data:{blabla}, msg: 'blalbla'}
	
	The keyword success is always present. It is mandatory for 
	the FormPanel widget of the extjs library, but not strictly 
	required since failure can be propagated through HTTP errors.
	
	In the data dictionary database field are encode as TableField.

	If the action failed an Exception is raised.
	The exception can be catch to generate an HTTP error.

	The most important method are DbSvc.proceed(), DbSvc.create(),
	DbSvc.destroy(), DbSvc.read(), DbSvc.select(), DbSvc.update()

	"""

	KEYWORDS = ['data',
			    'debug',
			    'dbFields',
				'dir',
				'groupby',
				'leftjoin',
				'limit',
				'orderby',
				'sort',
				'start',
				'table', 
				'where',
				'xaction']


	def _encode_query(self, li):
		""" helper method to encode the list of query as a web2py query.
		
				"[table1, field1] == [table2, field2]"
				"[table1, field1] > n"
				"[table1, field1] like 'm%'"
				"[table1, field1].lower() like 'm%'"
				"[table1, field1] belongs (1, 2, 3)"
				"([table1, field1] == [table2, field2]) | ([table1, field2] == [table2, field1])"
				
				all element of the list are ANDED
				operator are ==, !=, <, >, <=, >=
				boolean operator are &, |
		""" 
		query = ""
		for el in li:
			
			s = el.replace(' ','')
			s = s.replace(',', '.')
			s = s.replace('[', 'db.')
			s = s.replace(']', '')
			
			if 'belongs' in s:
				s = "%s)" % s.replace('belongs', '.belongs(' )
			elif 'like' in s:
				s = "%s)" % s.replace('like', '.like(' )
				
			query = ("(%s) & (%s)" % (query, s) if query else s)

		return eval(query, {}, {"db": self._db})	

	
	def _get_fields(self):
		"""Helper method returning a dictionary with request fields
		and values.
		
		The App.data.JsonStore send the field and their values
		as steering keywords for FormPanel. 
		
		Since EXtJs 3.1, the fields and their values are 
		encapsulated in the data dictionary for a GridPanel.
		
		This method allows to extract field values in all cases.
		
		"""
		# New approach coming with ExtJs 3.1 where fields values are
		# encapsulated in the data dictionary
		if 'data' in self._vars:
			return self._vars['data']
		
		# preserve backward compatibility for formPanel
		fields = dict(self._vars)
		for key in self.KEYWORDS:
			if key in fields:
				del fields[key]
		return fields
		
		
	def _is_action_valid(self, action):
		"""Raise an Exception if the request action is not handle."""
		if action not in ACTIONS:
			raise Exception, ACTION_INVALID % action


	def _is_field_in_table(self, table, field):
		""" Return true is the field belongs to the table. """
		return field in self._db[table].fields
	
	
	def _is_fields_values_valid(self, table, fields):
		"""Return a dictionary {field1: error1, field2: error2,...}
		
		The value of each field is check against validators defined in the
		data model. The return dictionary contains field failing these checks.
		
		the dictionary is empty when everything is OK.
		
		"""
		di = {}
		for (field, val) in fields.items():
			errmsg = self._db[table][field].validate(val)[1]
			if val and self._db[table][field].notnull and errmsg:
				di[field] = errmsg
		return di


	def _is_table_in_db(self, table):
		"""Raise an Exception if the table is not defined in the database."""
		if table not in self._db.tables:
			raise Exception, TABLE_NOT_IN_DB % table


	def create(self):
		"""Create a new record in a table.

			debug
				boolean

			table
				the table name where data will be inserted.

			data
				a dictionary with field values {field1: x, field2: y,...}
				approach use by App.data.JsonStore embedded in GridPanel
				since ExtJs 3.1
				
			field1, field2,... 
				To keep backward compatibility with formPanel
				the field value.

				
		Return a dictionary with status, error message and id of the insert row:
			{success: False, errors: {field: 'blabla',..}}
			The keyword success and error are required by 
			the Ext.FormPanel widget.
		
			or
			
			{success: True, msg: 'blalbla', data:{id:xx}}

		"""
		self._debug("Start DbSvc.insert")
		
		self._is_keyword("table")
		
		table = self._vars.table
		self._is_table_in_db(table)

		# Remove fields which do not belong to the table.
		# The JsonStore send the foreign key and the the pointing field
		# when manipulating table with foreign keys. The latter belongs
		# to another table. 
		
		fields, di = {}, self._get_fields()
		for table_field in di:
			tablename, fieldname = decode_field(table_field)
			if tablename == table:
				fields[fieldname.encode('utf8')] = di[table_field]

		# Validate field contents
		di = self._is_fields_values_valid(table, fields)
		if di:
			return {"success": False, "errors": di}
		
		# Create the new record
		self._debug('fields -->', fields)
		id = self._db[table].insert (**fields)

		self._debug("End DbSvc.insert.")
		resp = {"success": True, 
			    "msg": RECORD_INSERTED % id, 
			    "data": {encode_field(table,'id'): id}}
		return resp 


	def destroy(self):
		"""Destroy a record in a table.

			debug
				boolean

			table
				the table name where record will be removed.

			data
				dictionary with the identifier of the record (id)
				
		Return a dictionary with status and error message:
		{success: True, msg: 'blalbla'}

		"""
		self._debug("Start DbSvc.destroy")

		self._is_keyword("table")
		self._is_keyword("data")
		
		table = self._vars.table
		self._is_table_in_db(table)
		
		id = self._vars.data[encode_field(table, "id")]
		del self._db[table][id]
		
		self._debug("End DbSvc.destroy")
		return {"success": True, "msg": RECORD_DELETED % id}


	def proceed(self, vars):
		"""Proceed the request sends via the POST/GET method.

		the input vars is a Storage object.

		Return a dictionary with status, data and error message:
		{success: True, data:{blabla}, msg: 'blalbla'}

		"""
		BaseSvc.proceed(self, vars)
		self._debug("Start DbSvc.proceed")

		self._is_keyword("xaction")
		self._is_action_valid(vars.xaction)
		
		# create, destroy, update and read action related to a single table
		# they are used to manipulate record within a table
		if "table" in vars:
			self._debug("self.%s()" % vars.xaction)
			di = eval("self.%s()" % vars.xaction)

		# complex query involving several tables and conditions
		# they are use to build view or report
		elif vars.xaction == 'read':
			di = self.select()
			
		else:
			raise Exception, ACTION_INVALID % ""
		
		self._debug("End DbSvc.proceed")
		return di


	def read(self):
		"""Read the content of a single table resolving foreign keys.
		
			debug
				A boolean
			
			table
				name of the table containing records
				
			dbFields
				A list of field to be red.
				It also include pointing field resolving foreign keys.
				[(table1, field1), (table1, field2), (table2, field3), ...]
							
			where (optional)
				define the inner join when handling foreign key.
				The syntax follows the web2py convention:
				['db.table1.field1  == db.table2.id'].
				An AND is performed between the different 
				elements of the list.
				
				
		This method also understand parameters useful for paging.
		For more detail see  Ext.PagingToolbar.
			start
			limit
			sort
			dir
			
		Return a dictionary with selected data:
		{success: True, data: [{field1: xxx,..},...]}

		"""
		self._debug("Start DbSvc.read")
		self._is_keyword("dbFields")

		self._is_keyword("table")		
		table = self._vars.table
		self._is_table_in_db(table)
		
		# build the list of SQLField object
		# The JSON process sends a string when only one field is defined
		# NOTE: The json encoding return the tuple [table, field] 
		#       as a string 'table, field'
		fields = []
		for el in self._list(self._vars.dbFields):
			(table, field) = el.replace(' ','').split(',')
			if not self._is_field_in_table(table, field):
				raise Exception, FIELD_NOT_IN_DB % ("%s.%s" % (table, field))
			fields.append(self._db[table][field])

		# build the inner join conditions
		# The JSON process sends a string when only one field is defined
		query = ''
		if "where" in self._vars:
			query = self._encode_query(self._list(self._vars["where"]))
		
		# handle paging options (see Ext.PagingToolbar)
		# Count the number of record in the table --require for the paging
		kwargs, results = {}, None
		if 'start' in self._vars and 'limit' in self._vars:
			start = int(self._vars.start)
			limit = int(self._vars.limit)
			kwargs['limitby'] = (start, start+limit)
			results = self._db(self._db[table].id).count()
		
		rows = self._db(query).select(*fields, **kwargs)

		self._debug("End DbSvc.read")
		return {"success": True, "results": results, "data": rows_serializer(rows)}


	def select(self):
		""" Handle complex queries with inner/left join and with where
		group by, order by conditions. 

			debug
				A boolean
			
			dbFields
				A list of field to be red.
				[(table1, field1), (table2, field1), ....]
							
			where (optional)
				define the inner join when handling foreign key.
				The syntax follows the web2py convention:
				['db.table1.field1  == db.table2.id', 'db.table2.field3 > 3',...].
				An AND is performed between the different 
				elements of the list.
				
			left (optional)
			
			groupby (optional)
			
			orderby (optional)
			
			distinct (optional)
			
				
		This method also understand parameters useful for paging.
		For more detail see Ext.PagingToolbar.
			start
			limit
			sort
			dir
			
		Return a dictionary with selected data:
		{success: True, data: [{Table1Field1: xxx,..},...]}
		
		"""
		self._debug("Start DbSvc.read")
		self._is_keyword("dbFields")
		
		# build the list of SQLField object
		# The JSON process sends a string when only one field is defined
		# but a list in the other case.
		# NOTE: The json encoding return the tuple [table, field, func] 
		# as a string 'table, field, func'
		fields = []
		for el in self._list(self._vars.dbFields):
			li = el.replace(' ','').split(',')
			(table, field) = li[:2]
			if not self._is_field_in_table(table, field):
				raise Exception, FIELD_NOT_IN_DB % ("%s.%s" % (table, field))
			
			if len(li) == 2:
				fields.append(self._db[table][field])
				
			elif len(li) == 3:
				func = li[2]
				if func not in FUNC_FIELDS:
					raise Exception, FUNC_FIELD_INVALID % func
					
				dbfield = "db.%s.%s.%s()" % (table, field, func)
				fields.append(eval(dbfield, {}, {"db": self._db}))
			
		# build the innerJoins / where conditions
		# The JSON process sends a string when only one field is defined
		query = ''
		if "where" in self._vars:
			query = self._encode_query(self._list(self._vars["where"]))
					
		# handle left join
		kwargs = {}
		if "table" in self._vars and "left" in self._vars:
			table = self._vars.table
			self._is_table_in_db(table)
			left_query = self._encode_query(self._list(self._vars["left"]))
			kwargs["left"] = self._db[table].on(left_query)
			
		# handle orderby
		if "orderby" in self._vars:
			for el in self._list(self._vars.orderby):
				(table, field) = el.replace(' ','').split(',')
				if "orderby" in kwargs:
					kwargs["orderby"] = kwargs["orderby"] | self._db[table][field]
				else:
					kwargs["orderby"] = self._db[table][field]
						
		# handle group by
		if "groupby" in self._vars:
			(table, field) = self._vars.groupby.replace(' ','').split(',')
			kwargs["groupby"] = self._db[table][field]
		
		# handle distinct
		if "distinct" in self._vars:
			kwargs["distinct"] = True
			
		# handle paging options (see Ext.PagingToolbar)
		if 'start' in self._vars and 'limit' in self._vars:
			kwargs['limitby'] = (int(self._vars.start), int(self._vars.limit))
		
		# proceed the select
		rows = self._db(query).select(*fields, **kwargs)

		self._debug("End DbSvc.read")
		return {"success": True, "data": rows_serializer(rows)}
		
		
	def update(self):
		"""Update a record in a table.

		The steering keywords are:
		
			debug
				boolean
				
			table
				the table name where data will be updated
				
			data
				A dictionary containing the update values for field
				as well as the identifier of the record to be updated (id)
				

		Return a dictionary with status, message and id of the update row:
		{success: True, msg: 'blalbla', data:{id:xx}}

		"""
		self._debug("Start DbSvc.update.")
		
		self._is_keyword("table")
		self._is_keyword("data")
		
		table = self._vars.table
		self._is_table_in_db(table)

		# Remove fields which do not belong to the table.
		# The JsonStore send the foreign key and the the pointing field
		# when manipulating table with foreign keys. the latter belongs
		# to another table.
		
		fields = {}
		for table_field in self._vars.data:
			tablename, fieldname = decode_field(table_field)
			if tablename == table:
				fields[fieldname.encode('utf8')] = self._vars.data[table_field]

		# Validate field contents
		di = self._is_fields_values_valid(table, fields)
		if di:
			return {"success": False, "errors": di}
			
		id = fields["id"]
		self._db[table][id] = fields

		self._debug("End DbSvc.update.")
		return {"success": True, "msg": RECORD_UPDATED % id, "data": {"id": id}}
